/*
	Licensed Material - Property of IBM

	(C) Copyright 2015, 2016 IBM Corp.

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.
*/


'use strict';

var moduleName = 'services'; // - name for clarification during logging

var serverSupport = require('mdo-server-support');
var strings = require('ibm-strings');
var errors = require('mdo-errors');
var messages = require('./externalizedStrings');
var async = require('async');
var log = require('npmlog');
var Q = require('q');
var http = require('http');
var https = require('https');

const servicesMap = {
    push : getPush,
    liveupdate: getLiveUpdate,
}

module.exports.getServices = function(serverInfo, appName) {
  var deferred = Q.defer();

  var services = [];
  const keys = Object.keys(servicesMap);
  async.forEachSeries(keys, function (key, callback) {
    servicesMap[key](serverInfo, appName).then(function(service) {
      if (service && (Object.keys(service)).length > 0) {
        service.name = key;
        services.push(service);
      }
      callback();
    }).catch(function (error) {   
      callback(error);
    })    
  }, function(err) {
    if (err) {
      deferred.reject(err);
    } else {
      if (services.length > 0) {
        deferred.resolve(services);
      } else {
        deferred.resolve();
      }
    }
  });

  return deferred.promise;
}

function getPush(serverInfo, appName) {
  var deferred = Q.defer();

  var service = {};
  exportTags(`/mfpadmin/management-apis/2.0/runtimes/mfp/notifications/applications/${appName}/tags`, serverInfo, appName, []).then(function (tags) {
    if (tags && tags.length > 0) {
      service.tags = tags;
    }
    return getSettings(serverInfo, appName);
  }).then(function (settings) {
    if (settings && (Object.keys(settings)).length > 0 ) {
      service.settings = settings;
    }
    deferred.resolve(service);  
  }).catch(function (err) {
    deferred.resolve();
  })  
  
  return deferred.promise;                 
}

function exportTags(url, serverInfo, appName, tags) { 
  var deferred = Q.defer();

  getTags(url, serverInfo).then(function(response) {
    if (response && response.tags && response.tags.length > 0) {
      response.tags.forEach(function (responseTag) {
        var tag = {};
        tag.name = responseTag.name;
        tag.description = responseTag.description;
        tags.push(tag);
      });
    }     

    if (response && response.pageInfo && response.pageInfo.next) {
      var url = response.pageInfo.next;
      if (url.includes('?')) {
        var queryParams  = url.split('?')[1];
        url = `/mfpadmin/management-apis/2.0/runtimes/mfp/notifications/applications/${appName}/tags?` + queryParams;
      }
      exportTags(url, serverInfo, appName, tags).then(function(tags){
        deferred.resolve(tags);
      })
    } else {
        deferred.resolve(tags);
    }         
  }).catch(function (err) {
      deferred.resolve();       
  });  
  
  return deferred.promise;     
}

function getTags(url, serverInfo) {    
  var deferred = Q.defer();

  var options = {
    path : url,
    method : 'GET'
  }

  sendRequest(serverInfo, options).then(function (response) {
    deferred.resolve(response);                  
  }).catch(function (err) {
    deferred.resolve();
  });
  
  return deferred.promise;
}

function getSettings(serverInfo, appName) {
  var deferred = Q.defer();

  var settings = {};
  getGcmConf(serverInfo, appName).then(function(result) {
    if (result) {
      settings.gcm = result;
    }
    return getApnsConf(serverInfo, appName);
  }).then(function(result) {
    if (result) {
      settings.apns = result;
    }
    deferred.resolve(settings);
  }).catch(function (err) {
    deferred.resolve();
  }); 
  
  return deferred.promise;
}

function getGcmConf(serverInfo, appName) {
  var deferred = Q.defer();

  var options = {
    path : `/mfpadmin/management-apis/2.0/runtimes/mfp/notifications/applications/${appName}/gcmConf`,
    method : 'GET'
  }

  sendRequest(serverInfo, options).then(function (response) {
    if (response) {
      delete response.productVersion;
      deferred.resolve(response);
    } else {
      deferred.resolve();
    }
  }).catch(function (err) {
    deferred.resolve();
  });  
  
  return deferred.promise;
}

function getApnsConf(serverInfo, appName) {
  var deferred = Q.defer()

  var options = {
    path : `/mfpadmin/management-apis/2.0/runtimes/mfp/notifications/applications/${appName}/apnsConf`,
    method : 'GET'
  }

  sendRequest(serverInfo, options).then(function (response) {
    if (response) {
      delete response.productVersion;      
      delete response.validUntil;
      response.password = '<<Insert password here>>';
      deferred.resolve(response);
    } else {
      deferred.resolve();
    }
  }).catch(function (err) {
    deferred.resolve();
  });  
  
  return deferred.promise;
}

function getLiveUpdate(serverInfo, appName) {
  var deferred = Q.defer();

  var service = {};
  getSchema(serverInfo, appName).then(function (schema) {
    if (schema && (Object.keys(schema)).length > 0) {
      if ((Object.keys(schema.features)).length > 0 || (Object.keys(schema.properties)).length  > 0) {
        service.schema = schema;
      }
    }
    deferred.resolve(service);
  }).catch(function (err) {
    deferred.resolve();
  }) 

  return deferred.promise;                   
}

function getSchema(serverInfo, appName) {
  var deferred = Q.defer()

  var options = {
    path : `/mfpadmin/management-apis/2.0/runtimes/mfp/mfpliveupdate/${appName}/schema`,
    method : 'GET'
  }
  sendRequest(serverInfo, options).then(function (response) {
    if (response) {
      deferred.resolve(response);
    } else {
      deferred.resolve();
    }
  }).catch(function (err) {
    deferred.resolve();
  });  
  
  return deferred.promise;
}

function sendRequest(serverInfo, options) {
    var deferred = Q.defer();

    options.hostname = serverInfo.host;
    options.port = serverInfo.port;
    options.runtime = serverInfo.runtime;
    options.protocol = serverInfo.protocol + ':';

    options.headers = {
      'Content-Type' : 'application/json',
      'Authorization' : 'Basic ' + new Buffer(serverInfo.login + ':' + serverInfo.password).toString('base64'),
      'Accept-Charset':'UTF-8',
      'User-Agent':''
    };
    var requestCallback = function(response) {
        var content = '';
        var errs;
  
        if (response.statusCode === 401) {
          var invalidConfidentialClient = errors.getError('INVALIDCONFIDENTIALCLIENT');
          deferred.reject(invalidConfidentialClient);
        }
  
        response.on('data', function (data) {

          if (data) {
            try {
              content += decodeURI(data);
            } catch(err) {
              log.silly(moduleName, 'Failed to perform decodeURI: ' + err);
              content += data;
            }
            if (data.errors) errs = data.errors;
          }
        });
  
        response.on('end', function() {
          if (response.statusCode === 200 || response.statusCode === 201 ) {
            if (content !== '') {
              try {
                var results = JSON.parse(content);
                log.silly(moduleName, 'Results: ' + JSON.stringify(results, null, 2));
  
                if (results.errors && results.errors.length) {
                  strings.fatal(messages.logErrorLabel);
                  results.errors.forEach(function (error) {
                    strings.fatal(error);
                  });
                }
                if (results.warnings && results.warnings.length) {
                  strings.warn(messages.logWarningLabel);
                  results.warnings.forEach(function (warning) {
                    strings.warn(warning);
                  });
                }
                if (results.code && results.message) {
                  deferred.resolve();
                } else {
                  deferred.resolve(results);
                }
              } catch (e) {
                strings.log(content);
              }
            } else {
              log.silly(moduleName, 'Received empty response');
            }
          } else {
            log.silly(moduleName, (strings.format(messages.statusCodeMessage,response.statusCode,http.STATUS_CODES[response.statusCode]))[response.statusCode > 400 ? strings.theme.error : strings.theme.warn]);
            if (content !== '') log.silly(moduleName, JSON.parse(content));
          }
          deferred.resolve();
        });
    };
  
    var request;
    options.path = encodeURI(options.path);
 
    log.silly(moduleName, 'call push api:  \'%j\'', options);
    request = getRequest(options, requestCallback);    

    request.on('error', function (e) {
        // If there's an error with the request, log it and resolve the deferral
        if (e.code === 'ECONNREFUSED') {
          strings.fatal(e.message);
          deferred.reject(e.code);
        }
        else {
          strings.log(messages.errorInvoking);
          strings.log(e.message, strings.theme.error);
        }
        deferred.reject();
    });
    request.end();

    return deferred.promise;
}

function getRequest(options, callback) {
    serverSupport.getCertificate(options, moduleName);
  
    if (options.protocol === 'https:') {
      log.silly(moduleName, 'performing https request');
      return https.request(options, callback);
    } else {
      log.silly(moduleName, 'performing http request');
      return http.request(options, callback);
    }
  }
